#ifndef COMPONENT_H
#define COMPONENT_H
#include "..\includes.h"

class Component: public sc_module
{
public:
	sc_in<sc_logic> clock;
	sc_in<sc_logic> reset;
	sc_in<sc_logic> enable;

	SC_HAS_PROCESS(Component);
	Component() : clock("clock"), reset("reset"), enable("enable")
	{
		SC_CTHREAD(BehaviouralModel, this->clock.pos());
	}

	void BindCommonPorts(Component* componentToBind)	// Does not work if it is used in hierarchical components... The way of binding (at the first this or componentToBind affects on wich hierachical level port are binded
	{
		this->clock.bind(componentToBind->clock);
		this->reset.bind(componentToBind->reset);
		this->enable.bind(componentToBind->enable);
	}

	void BehaviouralModel()
	{
		while(1)
		{
			wait();
			if(this->reset.read() == SC_LOGIC_1)
			{
				this->Reset();
			}
			else if(this->enable.read() == SC_LOGIC_1)
			{
				this->Update();
			}
		}
	}

	virtual void Update() = 0;
	virtual void Reset() = 0;
	//void TraceSignals(sc_trace_file *tf);
};

#ifdef MTI_SYSTEMC
class VHDL_Component : public sc_foreign_module
{
public:
	sc_in<sc_logic> clock;
	sc_in<sc_logic> enable;
	sc_in<sc_logic> reset;

    VHDL_Component(sc_module_name name) : sc_foreign_module(name),
		clock("clock"), enable("enable"), reset("reset"){}
	void BindCommonPorts(Component* componentToBind)	
	{
		this->clock.bind(componentToBind->clock);
		this->reset.bind(componentToBind->reset);
		this->enable.bind(componentToBind->enable);
	}

	~VHDL_Component(){}
};
#endif //MTI_SYSTEMC

#endif //COMPONENT_H
