#ifndef CHECKNODE_H
#define CHECKNODE_H
#include "..\Component.h"
#include "Memory.h"
#include "VariableShiftRegister.h"

#ifndef MTI_SYSTEMC
#define CN_RAM_WIDTH 2*WIDTH_MAGNITUDE + DEGREE_WIDTH + MAX_DEGREE + 1
class CheckNodeFunctionalUnit : public Component
{
public:
#pragma region Interface
	sc_in<TYPE > in_input;
	sc_out<TYPE > output;
	// for the calculation of extrinsic messages
	sc_in<sc_lv<DEGREE_WIDTH> > in_indexIn;
	sc_in<sc_lv<DEGREE_WIDTH> > in_indexOut;
	sc_in<sc_logic> in_firstValue;						// true stands for being the first message of the check node
	sc_in<sc_logic> in_lastValue;
	sc_in<bool> in_firstIteration;
	sc_in<sc_lv<DEGREE_WIDTH> > in_variableDelay;

	sc_in<sc_logic> in_enableMEM;
	sc_in<sc_logic> in_readMEM;
	sc_in<sc_lv<RAM_CN_ADDRESS_WIDTH> > in_addressMEM;
	sc_out<sc_logic> out_valid;								// used to determine if the CheckNode thinks the decoded codeword has to be valid
#pragma endregion

	sc_signal<TYPE > input;
	sc_signal<sc_lv<DEGREE_WIDTH> > index_of_the_input_extrinsic_message;
#pragma region Memory
	SinglePortMemory<CN_PER_FU, CN_RAM_WIDTH, RAM_CN_ADDRESS_WIDTH>* extrinsicMemory_opt;
	sc_signal<sc_lv<CN_RAM_WIDTH> > memory_opt_input;
	sc_signal<sc_lv<CN_RAM_WIDTH> > memory_opt_output;
	sc_signal<TYPE > extrinsic_memory_output;
#pragma endregion

#pragma region "Registers for calculation"
	// registers for input values
	sc_signal<bool> sdSign;
	sc_signal<bool> sdSign_total;
	sc_signal<bool> sign_of_input_message;
	sc_signal<MAGNITUDE> magnitude_of_input_message;
	// registers for the calculation of extrinsic messages
	sc_signal<MAGNITUDE> min1;						// magnitude of the min
	sc_signal<MAGNITUDE> min2;						// second magnitude
	sc_signal<sc_lv<DEGREE_WIDTH> > index_min;						// absolute
	sc_signal<bool> sign_total;
	VariableShiftRegisterCN<bool>* sign_shift_register;
	// registers for saving the extrinsic messages
	sc_signal<MAGNITUDE> min1_calculated;			// magnitude of the min
	sc_signal<MAGNITUDE> min2_calculated;			// second magnitude
	sc_signal<sc_lv<DEGREE_WIDTH> > index_min_calculated;			// absolute
	sc_signal<bool> sign_total_calculated;
	sc_signal<bool> sign_of_output;
	sc_signal<sc_lv<DEGREE_WIDTH> > index_in_delayed;
	ShiftRegister<sc_lv<DEGREE_WIDTH> >* shift1;
	ShiftRegister<sc_lv<DEGREE_WIDTH> >* shift2;
	sc_signal<sc_lv<DEGREE_WIDTH> > sizeOfShiftRegister;
	sc_signal<sc_lv<MAX_DEGREE> > signs;
#pragma endregion

#pragma region "helper variable"
	//int index_of_the_input_extrinsic_message;
	MAGNITUDE max_magnitude;
	FILE* file_Extr;
#pragma endregion

	SC_HAS_PROCESS(CheckNodeFunctionalUnit);
	CheckNodeFunctionalUnit(sc_module_name name) :
		input("input"),
		output("output"),
		in_indexIn("in_indexIn"),
		in_indexOut("in_indexOut"),
		in_firstValue("in_firstValue"),					// true stands for being the first message of the check node
		in_lastValue("in_lastValue"),
		in_firstIteration("in_firstIteration"),
		in_variableDelay("in_variableDelay"),
		in_enableMEM("in_enableMEM"),
		in_readMEM("in_readMEM"),
		in_addressMEM("in_addressMEM"),
		out_valid("out_valid"),					// used to determine if the CheckNode thinks the decoded codeword has to be valid

		memory_opt_input("memory_opt_input"),
		memory_opt_output("memory_opt_output"),
		extrinsic_memory_output("extrinsic_memory_output"),
		sdSign("sdSign"),
		sdSign_total("sdSign_total"),
		sign_of_input_message("sign_of_input_message"),
		magnitude_of_input_message("magnitude_of_input_message"),
		min1("min1"),						// magnitude of the min
		min2("min2"),						// second magnitude
		index_min("index_min"),						// absolute
		sign_total("sign_total"),
		min1_calculated("min1_calculated"),			// magnitude of the min
		min2_calculated("min2_calculated"),			// second magnitude
		index_min_calculated("index_min_calculated"),			// absolute
		sign_total_calculated("sign_total_calculated"),
		sign_of_output("sign_of_output"),
		//index_in_delayed1("index_in_delayed1"),
		//index_in_delayed2("index_in_delayed2"),
		sizeOfShiftRegister("sizeOfShiftRegister")
	{
		SC_METHOD(MemoryInput);
		sensitive << this->sign_total_calculated << this->min1_calculated
			<< this->min2_calculated << this->index_min_calculated << this->signs;
		this->extrinsicMemory_opt = new SinglePortMemory<CN_PER_FU, CN_RAM_WIDTH, RAM_CN_ADDRESS_WIDTH>("ExtMemOpt");
		this->extrinsicMemory_opt->clock(this->clock);
		this->extrinsicMemory_opt->enable(this->in_enableMEM);
		this->extrinsicMemory_opt->reset(this->reset);
		this->extrinsicMemory_opt->address(this->in_addressMEM);
		this->extrinsicMemory_opt->readWriteEnable(this->in_readMEM);
		this->extrinsicMemory_opt->input(this->memory_opt_input);
		this->extrinsicMemory_opt->output(this->memory_opt_output);

		this->sign_shift_register = new VariableShiftRegisterCN<bool>("SignShift", MAX_DEGREE, 1, WIDTH_BOOL);
		this->sign_shift_register->BindCommonPorts(this);
		this->sign_shift_register->input(this->sign_of_input_message);
		this->sign_shift_register->output(this->sign_of_output);
		this->sign_shift_register->active_size(this->sizeOfShiftRegister);
#ifdef MTI_SYSTEMC
		this->sign_shift_register->read_enable(this->in_lastValue);
		this->sign_shift_register->content(this->signs);
#endif

		this->shift1 = new ShiftRegister<sc_lv<DEGREE_WIDTH> >("Shift1", 3, DEGREE_WIDTH);
		this->shift1->input(this->in_variableDelay);
		this->shift1->output(this->sizeOfShiftRegister);
		this->shift1->BindCommonPorts(this);

		this->shift2 = new ShiftRegister<sc_lv<DEGREE_WIDTH> >("Shift1", 3, DEGREE_WIDTH);
		this->shift2->input(this->in_indexIn);
		this->shift2->output(this->index_in_delayed);
		this->shift2->BindCommonPorts(this);

		this->max_magnitude = (unsigned int)((double)pow((double)2, WIDTH_MAGNITUDE) - 1);
		/*for( int i = 0; i < WIDTH_MAGNITUDE; i++)
		{
			this->max_magnitude[i] = SC_LOGIC_1;
		}*/
				// Debug
		char filename[200];
		strcpy (filename, "..\\Matlab\\");
		strcat (filename, this->name());
		strcat (filename, ".extr");
#ifdef MTI_SYSTEMC
		this->file_SDs = fopen(filename,"w");
#else
		if (fopen_s(&this->file_Extr, filename,"w") != NULL) 
		{ 
			printf("Failed to create file to write Extrinsic Messages.\n"); 
		}
#endif //MTI_SYSTEMC
	}

	void Reset()
	{
		this->min2.write(this->max_magnitude);		// Ones in every FlipFlop (max value)
		this->min1.write(this->max_magnitude);		// Ones in every FlipFlop (max value)
		this->min2_calculated.write(0);
		this->min1_calculated.write(0);
		this->magnitude_of_input_message.write(0);
	}

	void Update()
	{
		this->input.write(this->in_input.read());
		// read extrinsic from memory
		TYPE temp;
		sc_lv<CN_RAM_WIDTH> temp_row= this->memory_opt_output.read();
		this->index_of_the_input_extrinsic_message = this->in_variableDelay.read().to_uint() - this->in_indexIn.read().to_uint();
		if(temp_row.range(CN_RAM_WIDTH - 2 * WIDTH_MAGNITUDE - 1, CN_RAM_WIDTH - 2 * WIDTH_MAGNITUDE - DEGREE_WIDTH) == this->index_of_the_input_extrinsic_message.read())
		{
			temp = temp_row.range(CN_RAM_WIDTH - WIDTH_MAGNITUDE - 1, CN_RAM_WIDTH - 2 * WIDTH_MAGNITUDE).to_uint();
		}
		else
		{
			temp = temp_row.range(CN_RAM_WIDTH - 1, CN_RAM_WIDTH - WIDTH_MAGNITUDE).to_uint();
		}
		this->extrinsic_memory_output.write(Magnitude_to_SD(temp, temp_row[0] != temp_row[this->index_of_the_input_extrinsic_message.read().to_uint() + 1]));

		// calculate extrinsic input messages
		temp = this->input.read();
#ifdef DEBUG_SYSTEMC
		sc_fixed<WIDTH_SD, WIDTH_SD - 5> fixedPoint;
		fixedPoint.range(WIDTH_SD - 1, 0) = temp;
		double temp_double = fixedPoint.to_double();
#endif //DEBUG_SYSTEMC

		this->sdSign.write(Sign_of_SD(temp));
		bool temp_sign = Sign_of_SD(temp);
		if(this->in_firstIteration.read() == false)
		{
			TYPE temp2 = this->extrinsic_memory_output.read();
			temp = Subtract_SDs(temp, temp2);
		}
		this->magnitude_of_input_message.write(Magnitude_of_SD(temp));
		this->sign_of_input_message.write(Sign_of_SD(temp));

		// extrinsic message output
		MAGNITUDE min = this->index_min_calculated.read() == this->in_indexOut.read() ? this->min2_calculated.read() : this->min1_calculated.read();
		this->output.write(Magnitude_to_SD(min, sign_total_calculated.read() != this->sign_of_output.read()));

		// min1 / min2
		bool EN1 = Magnitude_Smaller(this->magnitude_of_input_message.read(), this->min1.read());
		bool EN2 = Magnitude_Smaller(this->magnitude_of_input_message.read(), this->min2.read());
		if(EN1 == true)
		{
			this->min2.write(this->min1.read());
		}
		else if(EN2 == true)
		{
			this->min2.write(this->magnitude_of_input_message.read());
		}
		if(this->in_firstValue.read() == SC_LOGIC_1)
		{
			this->min2.write(this->max_magnitude);
		}
		if(this->in_firstValue.read() == SC_LOGIC_1 || EN1)
		{
			this->min1.write(this->magnitude_of_input_message.read());
			this->index_min.write(this->index_in_delayed.read());
		}

		// if this is first value
		if(this->in_firstValue.read() == SC_LOGIC_1)
		{
			bool temp_sign = sdSign.read();
			this->sdSign_total.write(temp_sign);
			temp_sign = this->sign_of_input_message.read();
			this->sign_total.write(temp_sign);
		}
		else if(this->in_firstValue.read() == SC_LOGIC_0)
		{
			bool temp_sign = this->sdSign_total.read() != sdSign.read();
			this->sdSign_total.write(temp_sign);
			this->sign_total.write(this->sign_total.read() != this->sign_of_input_message.read());
		}

		// if this is last value
		if(this->in_lastValue.read() == SC_LOGIC_1)
		{
			bool temp_sign = this->sign_total.read();
			this->sign_total_calculated.write(temp_sign);
			this->min1_calculated.write(this->min1.read());
			this->min2_calculated.write(this->min2.read());
			this->index_min_calculated.write(this->index_min.read());
#ifndef MTI_SYSTEMC
			sc_lv<MAX_DEGREE> signs;
			for(int i = 0; i < MAX_DEGREE; i++)
			{
				signs[i] = this->sign_shift_register->shift_register[(MAX_DEGREE - 1) - i];
			}
			this->signs.write(signs);
#endif
			temp_sign = this->sdSign_total.read();
			if(!temp_sign)
			{
				this->out_valid.write(SC_LOGIC_1);
			}
			else
			{
				this->out_valid.write(SC_LOGIC_0);
			}
		}
	}

	void MemoryInput()
	{
		this->memory_opt_input.write((this->min1_calculated.read(), this->min2_calculated.read(), 
			this->index_min_calculated.read(), this->signs.read(), this->sign_total_calculated.read()));
	}

	void Print()
	{
		#ifndef MTI_SYSTEMC
		int temp = this->extrinsicMemory_opt->memory[0].read().to_uint();
		fprintf(file_Extr, "%d", temp);
		for(int i = 1; i < CN_PER_FU; ++i)
		{
			temp = this->extrinsicMemory_opt->memory[i].read().to_uint();
			fprintf(file_Extr, ",%d", temp);
 		}
		fprintf(file_Extr, "\n");
#endif //MTI_SYSTEMC
	}
};
#else
class CheckNodeFunctionalUnit : public VHDL_Component
{
	public:
	sc_in<TYPE > in_input;
	sc_out<TYPE > output;
	// for the calculation of extrinsic messages
	sc_in<sc_lv<DEGREE_WIDTH> > in_indexIn;
	sc_in<sc_lv<DEGREE_WIDTH> > in_indexOut;
	sc_in<sc_logic> in_firstValue;						// true stands for being the first message of the check node
	sc_in<sc_logic> in_lastValue;
	sc_in<bool> in_firstIteration;
	sc_in<sc_lv<DEGREE_WIDTH> > in_variableDelay;

	sc_in<sc_logic> in_enableMEM;
	sc_in<sc_logic> in_readMEM;
	sc_in<sc_lv<RAM_CN_ADDRESS_WIDTH> > in_addressMEM;
	sc_out<sc_logic> out_valid;


    CheckNodeFunctionalUnit(sc_module_name nm) :
		VHDL_Component(nm),
		in_input("in_input"),
		output("out_output"),
		in_indexIn("in_indexIn"),
		in_indexOut("in_indexOut"),
		in_firstValue("in_firstValue"),						// true stands for being the first message of the check node
		in_lastValue("in_lastValue"),
		in_firstIteration("in_firstIteration"),
		in_variableDelay("in_variableDelay"),

		in_enableMEM("in_enableMEM"),
		in_readMEM("in_readMEM"),
		in_addressMEM("in_addressMEM"),
		out_valid("out_valid")
    {
		elaborate_foreign_module("work.CheckNode");
    }
};
#endif //MTI_SYSTEMC
#endif //CHECKNODE_H
