#ifndef SYMBOLNODEFUNCTIONALUNIT_H
#define SYMBOLNODEFUNCTIONALUNIT_H
#include "..\..\..\ControlInformation\widths.h"
#include "..\Component.h"
#include "Memory.h"
#include "VariableShiftRegister.h"

//#ifndef MTI_SYSTEMC
class SymbolNodeFunctionalUnit : public Component
{
public:
	// Interface
	sc_in<TYPE > in_input;
	sc_out<TYPE > out_output;
	sc_in<TYPE > in_SD_read;
	sc_out<TYPE > out_SD_read;
	sc_in<TYPE > in_SD_write;
	sc_out<TYPE > out_SD_write;
	// Control Interface
	sc_in<sc_logic> in_initialization;
	sc_in<sc_logic> in_save_values;
	sc_in<sc_logic> in_oddIteration;
	sc_in<sc_logic> in_readEnable;
	sc_in<sc_logic> in_writeEnable;

	sc_in<sc_logic> in_first_extrinsic;
	sc_in<sc_logic> in_last_extrinsic;
	sc_in<sc_lv<RAM_SN_ADDRESS_WIDTH> > in_oldSD_address;
	sc_in<sc_lv<RAM_SN_ADDRESS_WIDTH> > in_newSD_address;


	sc_signal<TYPE > input;
	sc_signal<sc_logic> save_values;
	sc_signal<sc_logic> oddIteration;		
	sc_signal<sc_logic> addIntrinsic;
	sc_signal<sc_logic> resetSum;
	sc_signal<TYPE > partialSum;
	sc_signal<TYPE > partialSum2;
	sc_signal<TYPE > toAddTo_inSD_Write;
	// Data Memory
	SinglePortMemory<SN_PER_FU, WIDTH_SD, RAM_SN_ADDRESS_WIDTH>* intrinsicMemory;
	sc_signal<sc_lv<RAM_SN_ADDRESS_WIDTH> > intrinsicMemory_address;
	sc_signal<sc_logic> intrinsicMemory_readEnable;
	sc_signal<TYPE > intrinsicMemory_output;

	DualPortMemory<SN_PER_FU, WIDTH_SD, RAM_SN_ADDRESS_WIDTH>* sdMemory1;
	sc_signal<TYPE > sdMemory1_output;
	sc_signal<TYPE > sdMemory1_input;
	sc_signal<sc_logic> sdMemory1_readEnable;
	sc_signal<sc_logic> sdMemory1_writeEnable;
	sc_signal<sc_logic> sdMemory1_writeEnable_final;
	sc_signal<sc_lv<RAM_SN_ADDRESS_WIDTH> > sdMemory1_readAddress;
	sc_signal<sc_lv<RAM_SN_ADDRESS_WIDTH> > sdMemory1_writeAddress;
	sc_signal<sc_lv<RAM_SN_ADDRESS_WIDTH> > sdMemory1_writeAddress_final;

	DualPortMemory<SN_PER_FU, WIDTH_SD, RAM_SN_ADDRESS_WIDTH>* sdMemory2;
	sc_signal<TYPE > sdMemory2_output;
	sc_signal<TYPE > sdMemory2_input;
	sc_signal<sc_logic> sdMemory2_readEnable;
	sc_signal<sc_logic> sdMemory2_readEnable_final;
	sc_signal<sc_logic> sdMemory2_writeEnable;
	sc_signal<sc_lv<RAM_SN_ADDRESS_WIDTH> > sdMemory2_readAddress;
	sc_signal<sc_lv<RAM_SN_ADDRESS_WIDTH> > sdMemory2_readAddress_final;
	sc_signal<sc_lv<RAM_SN_ADDRESS_WIDTH> > sdMemory2_writeAddress;


	// Delays
	ShiftRegister<sc_logic>* shift1;
	ShiftRegister<sc_logic>* shift2;
	ShiftRegister<sc_logic>* shift3;
	ShiftRegister<sc_lv<RAM_SN_ADDRESS_WIDTH> >* shift4;
	ShiftRegister<TYPE >* shift6;
	ShiftRegister<sc_logic>* shift8;
	ShiftRegister<sc_lv<RAM_SN_ADDRESS_WIDTH> >* shift9;
	ShiftRegister<sc_logic>* shift10;
	// Debug
	FILE* file_SDs;

	SC_HAS_PROCESS(SymbolNodeFunctionalUnit);
	SymbolNodeFunctionalUnit(sc_module_name name, int index) :
		in_input("in_input"),
		out_output("out_output"),
		in_SD_read("in_SD_read"),
		out_SD_read("out_SD_read"),
		in_SD_write("in_SD_write"),
		out_SD_write("out_SD_write"),
		in_initialization("in_initialization"),
		in_save_values("in_save_values"),
		in_oddIteration("in_oddIteration"),
		in_readEnable("in_readEnable"),
		in_writeEnable("in_writeEnable"),
		//in_address("in_address"),
		input("input"),
		//oldSD_Address("oldSD_Address"),
		//firstExtrinsicMessage("firstExtrinsicMessage"),
		addIntrinsic("addIntrinsic"),
		//lastExtrinsicMessage("lastExtrinsicMessage"),
		resetSum("resetSum"),	
		intrinsicMemory_address("intrinsicMemory_address"),
		intrinsicMemory_readEnable("intrinsicMemory_readEnable"),
		//sc_signal<TYPE > intrinsicMemory_input;
		intrinsicMemory_output("intrinsicMemory_output"),
		sdMemory1_output("sdMemory1_output"),
		sdMemory1_input("sdMemory1_input"),
		sdMemory1_readEnable("sdMemory1_readEnable"),
		sdMemory1_writeEnable("sdMemory1_writeEnable"),
		sdMemory1_readAddress("sdMemory1_readAddress"),
		sdMemory1_writeAddress("sdMemory1_writeAddress"),
		sdMemory1_writeAddress_final("sdMemory1_writeAddress_final"),
		sdMemory2_output("sdMemory2_output"),
		sdMemory2_input("sdMemory2_input"),
		sdMemory2_readEnable("sdMemory2_readEnable"),
		sdMemory2_writeEnable("sdMemory2_writeEnable"),
		sdMemory2_readAddress("sdMemory2_readAddress"),
		sdMemory2_readAddress_final("sdMemory2_readAddress_final"),
		sdMemory2_writeAddress("sdMemory2_writeAddress"),
		in_newSD_address("in_newSD_address")
	{
		SC_METHOD(SdMemory2_Access);
		sensitive << this->in_oddIteration << this->in_oldSD_address << this->in_newSD_address << this->in_readEnable << this->in_writeEnable;
		SC_METHOD(PartialSum);
		sensitive << this->input << this->addIntrinsic << this->intrinsicMemory_output << this->sdMemory2_output;

		this->intrinsicMemory = new SinglePortMemory<SN_PER_FU, WIDTH_SD, RAM_SN_ADDRESS_WIDTH>("MEM_Intr");
		this->intrinsicMemory->BindCommonPorts(this);
		this->intrinsicMemory->address(this->intrinsicMemory_address);
		this->intrinsicMemory->readWriteEnable(this->intrinsicMemory_readEnable);
		this->intrinsicMemory->input(this->sdMemory1_input);
		this->intrinsicMemory->output(this->intrinsicMemory_output);

		this->sdMemory1 = new DualPortMemory<SN_PER_FU, WIDTH_SD, RAM_SN_ADDRESS_WIDTH>("MEM_SD1");
		this->sdMemory1->BindCommonPorts(this);
		this->sdMemory1->readEnable(this->sdMemory1_readEnable);
		this->sdMemory1->readAddress(this->sdMemory1_readAddress);
		this->sdMemory1->writeEnable(this->sdMemory1_writeEnable_final);
		this->sdMemory1->writeAddress(this->sdMemory1_writeAddress_final);
		this->sdMemory1->input(this->sdMemory1_input);
		this->sdMemory1->output(this->sdMemory1_output);

		this->sdMemory2 = new DualPortMemory<SN_PER_FU, WIDTH_SD, RAM_SN_ADDRESS_WIDTH>("MEM_SD2");
		this->sdMemory2->BindCommonPorts(this);
		this->sdMemory2->readEnable(this->sdMemory2_readEnable_final);
		this->sdMemory2->readAddress(this->sdMemory2_readAddress_final);
		this->sdMemory2->writeEnable(this->sdMemory2_writeEnable);
		this->sdMemory2->writeAddress(this->sdMemory2_writeAddress);
		this->sdMemory2->input(this->sdMemory2_input);
		this->sdMemory2->output(this->sdMemory2_output);

		// Delay between Read and Write of SD-Values
		this->shift2 = new ShiftRegister<sc_logic>("Shift2", DELAY_READ_WRITE - 1, WIDTH_BOOL);
		this->shift2->BindCommonPorts(this);
		this->shift2->input(this->sdMemory1_readEnable);
		this->shift2->output(this->sdMemory1_writeEnable);
		this->shift4 = new ShiftRegister<sc_lv<RAM_SN_ADDRESS_WIDTH> >("Shift4", DELAY_READ_WRITE - 1, RAM_SN_ADDRESS_WIDTH);
		this->shift4->BindCommonPorts(this);
		this->shift4->input(this->sdMemory1_readAddress);
		this->shift4->output(this->sdMemory1_writeAddress);
		this->shift8 = new ShiftRegister<sc_logic>("Shift8", DELAY_READ_WRITE, WIDTH_BOOL);
		this->shift8->BindCommonPorts(this);
		this->shift8->input(this->sdMemory2_readEnable_final);
		this->shift8->output(this->sdMemory2_writeEnable);
		this->shift9 = new ShiftRegister<sc_lv<RAM_SN_ADDRESS_WIDTH> >("Shift9", DELAY_READ_WRITE, RAM_SN_ADDRESS_WIDTH);
		this->shift9->BindCommonPorts(this);
		this->shift9->input(this->sdMemory2_readAddress_final);
		this->shift9->output(this->sdMemory2_writeAddress);

		// Delay due to different scheduling in odd and even iterations
		int delay = DURATION_PERMUTATION;		// +1 for the output and -1 for register toAddTo_inSD_Write
		this->shift6 = new ShiftRegister<TYPE >("Shift6", delay, WIDTH_SD);
		this->shift6->BindCommonPorts(this);
		this->shift6->input(this->partialSum2);
		this->shift6->output(this->toAddTo_inSD_Write);



		this->shift1 = new ShiftRegister<sc_logic>("Shift1", 1 + DURATION_RAM_ACCESS, WIDTH_BOOL);
		this->shift1->BindCommonPorts(this);
		this->shift1->input(this->in_first_extrinsic);
		this->shift1->output(this->addIntrinsic);
		this->shift3 = new ShiftRegister<sc_logic>("Shift3", 1 + DURATION_RAM_ACCESS, WIDTH_BOOL);
		this->shift3->BindCommonPorts(this);
		this->shift3->input(this->in_last_extrinsic);
		this->shift3->output(this->resetSum);
		this->shift10 = new ShiftRegister<sc_logic>("Shift10", 1 + DURATION_RAM_ACCESS, WIDTH_BOOL);
		this->shift10->BindCommonPorts(this);
		this->shift10->input(this->in_save_values);
		this->shift10->output(this->save_values);

		// Debug
		char filename[200];
		strcpy (filename, "..\\Matlab\\");
		strcat (filename, this->name());
		strcat (filename, ".sds");
#ifdef MTI_SYSTEMC
		this->file_SDs = fopen(filename,"w");
#else
		if (fopen_s(&this->file_SDs, filename,"w") != NULL) 
		{ 
			printf("Failed to create file to write SDs.\n"); 
		}
#endif //MTI_SYSTEMC
	}

	void Print()
	{
#ifndef MTI_SYSTEMC
		if(this->in_oddIteration.read() == SC_LOGIC_1)
		{
			int temp = this->sdMemory1->memory[0].read().to_int();
			fprintf(file_SDs, "%d", temp);
			for(int i = 1; i < SN_PER_FU; ++i)
			{
				temp = this->sdMemory1->memory[i].read().to_int();
				fprintf(file_SDs, ",%d", temp);
 			}
			fprintf(file_SDs, "%c", '\n');
		}
		else
		{
			int temp = this->sdMemory2->memory[0].read().to_int();
			fprintf(file_SDs, "%d", temp);
			for(int i = 1; i < SN_PER_FU; ++i)
			{
				temp = this->sdMemory2->memory[i].read().to_int();
				fprintf(file_SDs, ",%d", temp);
 			}
			fprintf(file_SDs, "%c", '\n');
		}
#endif //MTI_SYSTEMC
	}

	void SdMemory2_Access()
	{
		if(this->oddIteration.read() == SC_LOGIC_0)
		{
			this->sdMemory2_readEnable.write(this->in_writeEnable.read());
			this->sdMemory2_readAddress.write(this->in_newSD_address.read());
		}
		else
		{
			this->sdMemory2_readEnable.write(this->in_readEnable.read());
			this->sdMemory2_readAddress.write(this->in_oldSD_address.read());
		}
	}

	void PartialSum()
	{
		if(this->addIntrinsic.read() == SC_LOGIC_1)
		{
			this->partialSum.write(Add_SDs(this->input.read(), this->intrinsicMemory_output.read()));
		}
		else
		{
			this->partialSum.write(Add_SDs(this->input.read(), this->sdMemory2_output.read()));
		}
	}

	void Update()
	{
		// delay
		this->oddIteration.write(this->in_oddIteration.read());
		sc_logic oddIteration = this->oddIteration.read();
		this->input.write(this->in_input.read());

		// Intrinsic Memory Access
		sc_logic temp = this->in_initialization.read();
		this->intrinsicMemory_address.write(temp == SC_LOGIC_0 ? this->in_newSD_address.read() : this->in_oldSD_address.read());
		this->intrinsicMemory_readEnable.write(~temp);

		// SD1-Memory Access
		if(this->in_initialization.read() == SC_LOGIC_1)
		{
			this->sdMemory1_writeAddress_final.write(this->in_oldSD_address.read());
			this->sdMemory1_writeEnable_final.write(this->in_initialization.read());
			this->sdMemory1_input.write(this->in_input.read());
		}
		else
		{
			this->sdMemory1_writeAddress_final.write(this->sdMemory1_writeAddress.read());
			this->sdMemory1_writeEnable_final.write(this->sdMemory1_writeEnable.read());
			if(oddIteration == SC_LOGIC_0)
			{
				this->sdMemory1_input.write(this->in_SD_read.read());
				this->sdMemory1_readEnable.write(this->in_readEnable.read());
				this->sdMemory1_readAddress.write(this->in_oldSD_address.read());
			}
			else
			{
				this->sdMemory1_input.write(Add_SDs(this->in_SD_write.read(), this->toAddTo_inSD_Write.read()));
				this->sdMemory1_readEnable.write(this->in_writeEnable.read());
				this->sdMemory1_readAddress.write(this->in_newSD_address.read());
			}
		}

		// SD2-Memory Access
		if(this->in_save_values.read() == SC_LOGIC_1)
		{
			this->sdMemory2_readAddress_final.write(this->in_oldSD_address.read());
			this->sdMemory2_readEnable_final.write(this->in_save_values.read());
		}
		else
		{
			this->sdMemory2_readAddress_final.write(this->sdMemory2_readAddress.read());
			this->sdMemory2_readEnable_final.write(this->sdMemory2_readEnable.read());
		}
		if(oddIteration == SC_LOGIC_0)
		{
			this->sdMemory2_input.write(this->in_SD_write.read());
		}
		else
		{
			this->sdMemory2_input.write(this->in_SD_read.read());
		}

		// Output Logic to Pi2 Network and output
		this->out_output.write(0);
		this->out_SD_read.write(0);
		if(this->save_values.read() == SC_LOGIC_1)
		{
			this->out_output.write(this->sdMemory2_output.read());
		}
		else
		{
			this->out_output.write(oddIteration == SC_LOGIC_0 ? sdMemory1_output.read() : sdMemory2_output.read());
			this->out_SD_read.write(oddIteration == SC_LOGIC_0 ? sdMemory1_output.read() : sdMemory2_output.read());
		}
#ifdef DEBUG_SYSTEMC
		sc_fixed<WIDTH_SD, WIDTH_SD - 5> fixedPoint;
		fixedPoint.range(WIDTH_SD - 1, 0) = this->out_output.read();
		double temp_double = fixedPoint.to_double();
#endif //DEBUG_SYSTEMC
		
		// Output Logic to Pi3 Network
		if(oddIteration == SC_LOGIC_0) // even
		{
			this->out_SD_write.write(this->partialSum.read());
		}
		else
		{
			if(this->resetSum.read() == SC_LOGIC_0)
			{
				this->out_SD_write.write(this->sdMemory1_output.read());
			}
			else
			{
				this->out_SD_write.write(0);
			}
		}
			
		// Values to add to Input of Pi3
		this->partialSum2.write(this->addIntrinsic.read() == SC_LOGIC_1 ? this->partialSum.read() : this->input.read());
	}

	void Reset()
	{
		out_output.write(0);
		out_SD_read.write(0);
		out_SD_write.write(0);
	}
};
/*#else
class SymbolNodeFunctionalUnit : public VHDL_Component
{
	public:
	sc_in<TYPE > in_input;
	sc_out<TYPE > out_output;
	sc_in<TYPE > in_SD_read;
	sc_out<TYPE > out_SD_read;
	sc_in<TYPE > in_SD_write;
	sc_out<TYPE > out_SD_write;
	// Control Interface
	sc_in<sc_logic> in_initialization;
	sc_in<sc_logic> in_save_values;
	sc_in<sc_logic> in_oddIteration;
	sc_in<sc_logic> in_readEnable;
	sc_in<sc_logic> in_writeEnable;

	sc_in<sc_logic> in_first_extrinsic;
	sc_in<sc_logic> in_last_extrinsic;
	sc_in<sc_lv<RAM_SN_ADDRESS_WIDTH> > in_oldSD_address;
	sc_in<sc_lv<RAM_SN_ADDRESS_WIDTH> > in_newSD_address;


    SymbolNodeFunctionalUnit(sc_module_name nm, int bank) :
		VHDL_Component(nm),
		in_input("in_input"),
		out_output("out_output"),
		in_SD_read("in_SD_read"),
		out_SD_read("out_SD_read"),
		in_SD_write("in_SD_write"),
		out_SD_write("out_SD_write"),
		// Control Interface
		in_initialization("in_initialization"),
		in_save_values("in_saveValues"),
		in_oddIteration("in_oddIteration"),
		in_readEnable("in_readEnable"),
		in_writeEnable("in_writeEnable"),

		in_first_extrinsic("in_firstExtrinsic"),
		in_last_extrinsic("in_lastExtrinsic"),
		in_oldSD_address("in_oldSD_address"),
		in_newSD_address("in_newSD_address")
    {
		elaborate_foreign_module("work.SymbolNode");
    }
};
#endif*/
#endif // SYMBOLNODEFUNCTIONALUNIT_H
