#ifndef ACCESSROM_H
#define ACCESSROM_H
#include "..\Component.h"
#include "..\Components\Memory.h"
#include "..\Components\VariableShiftRegister.h"

#ifndef MTI_SYSTEMC
class ControlSNandPI : public Component
{
public:
	sc_in<sc_logic>										in_oddIteration;
	sc_in<sc_logic>										in_enableRom;
	sc_in<sc_lv<ROM_ADDRESS_WIDTH> >					in_addressRom;
	sc_in<sc_lv<DEGREE_WIDTH> >							in_variableDelay;

	sc_out<sc_logic>									out_newSD_readEnable;
	sc_out<sc_logic>									out_oldSD_readEnable;
	sc_out<sc_lv<(ROM_SN_WIDTH - 2) * PARALLELISM> >	out_oldSD_address_vector;
	sc_out<sc_lv<ROM_SN_WIDTH * PARALLELISM> >			out_newSD_address_vector;
	sc_out<sc_lv<4 * ROM_PERMUTATION_WIDTH> >			out_permutationIndices_vector;

	// SN
	RomMemory2<ROM_SIZE, ROM_SN_WIDTH * PARALLELISM, ROM_ADDRESS_WIDTH>* romSN;
	sc_signal<sc_lv<ROM_SN_WIDTH * PARALLELISM> > romSN_output;
	VariableShiftRegister<sc_lv<ROM_SN_WIDTH * PARALLELISM> >* variableShiftRomSN;
	VariableShiftRegister<sc_logic>* variableShift3;
	ShiftRegister<sc_logic>* delay_sn_read;
	ShiftRegister<sc_lv<DEGREE_WIDTH> >*				delay_sn_variableDelay;
	sc_signal<sc_lv<DEGREE_WIDTH> >							sn_variableDelay;

	// Pi
	sc_signal<sc_lv<DEGREE_WIDTH> > pi3_variableDelay;
	RomMemory2<ROM_SIZE, ROM_PERMUTATION_WIDTH, ROM_ADDRESS_WIDTH>* nodes_permutation_MEM;
	sc_signal<sc_logic> nodes_permutation_MEM_enable;
	sc_signal<sc_lv<ROM_ADDRESS_WIDTH> > permutation_index;
	VariableShiftRegister<sc_lv<ROM_PERMUTATION_WIDTH> >* variableShift_nodes_permutation;
	RomMemory2<ROM_SIZE, ROM_PERMUTATION_WIDTH, ROM_ADDRESS_WIDTH>* banks_permutation_MEM;
	sc_signal<sc_logic> banks_permutation_MEM_enable;
	sc_signal<sc_lv<ROM_ADDRESS_WIDTH> > permutation_index_delayed;	
	VariableShiftRegister<sc_lv<ROM_PERMUTATION_WIDTH> >* variableShift_banks_permutation;
	ShiftRegister<sc_lv<ROM_ADDRESS_WIDTH> >* shift5;
	ShiftRegister<sc_lv<ROM_ADDRESS_WIDTH> >* shift6;
	ShiftRegister<sc_logic>* shift10;
	ShiftRegister<sc_logic>* shift11;
	ShiftRegister<sc_lv<DEGREE_WIDTH> >* shift9;
	sc_signal<sc_lv<DEGREE_WIDTH> > pi1_variableDelay;
	ShiftRegister<sc_lv<DEGREE_WIDTH> >* shift8;
	sc_signal<sc_lv<ROM_PERMUTATION_WIDTH> > permutationIndices_temp[2];
	sc_signal<sc_lv<ROM_PERMUTATION_WIDTH> > permutationIndices[4];

	SC_HAS_PROCESS(ControlSNandPI);
	ControlSNandPI(sc_module_name name)
	{
		SC_METHOD(RomSNOutput);
		sensitive << this->romSN_output;
		SC_METHOD(PermutationOutput);
		for(int i = 0; i < 4; i++)
		{
			sensitive << this->permutationIndices[i];
		}

		this->romSN = new RomMemory2<ROM_SIZE, ROM_SN_WIDTH * PARALLELISM, ROM_ADDRESS_WIDTH>("RomSn");
		this->romSN->BindCommonPorts(this);
		this->romSN->readEnable(this->in_enableRom);
		this->romSN->readAddress(this->in_addressRom);
		this->romSN->output(this->romSN_output);

		this->delay_sn_variableDelay = new ShiftRegister<sc_lv<DEGREE_WIDTH> >("shift9", DURATION_ROM_ACCESS - 1, DEGREE_WIDTH);
		this->delay_sn_variableDelay->input(this->in_variableDelay);
		this->delay_sn_variableDelay->output(this->sn_variableDelay);
		this->delay_sn_variableDelay->BindCommonPorts(this);

		this->variableShiftRomSN = new VariableShiftRegister<sc_lv<ROM_SN_WIDTH * PARALLELISM> >("variableShiftRomSN", MAX_DEGREE, DELAY_NEW_SD, ROM_SN_WIDTH * PARALLELISM);
		this->variableShiftRomSN->BindCommonPorts(this);
		this->variableShiftRomSN->input(this->romSN_output);
		this->variableShiftRomSN->output(this->out_newSD_address_vector);
		this->variableShiftRomSN->active_size(this->sn_variableDelay);

		this->delay_sn_read = new ShiftRegister<sc_logic >("DelaySnRead", DURATION_ROM_ACCESS, WIDTH_BOOL);
		this->delay_sn_read->input(this->in_enableRom);
		this->delay_sn_read->output(this->out_oldSD_readEnable);
		this->delay_sn_read->BindCommonPorts(this);

		this->variableShift3 = new VariableShiftRegister<sc_logic>("VariableShift3", MAX_DEGREE, DELAY_NEW_SD, 1);
		this->variableShift3->BindCommonPorts(this);
		this->variableShift3->input(this->out_oldSD_readEnable);
		this->variableShift3->output(this->out_newSD_readEnable);
		this->variableShift3->active_size(this->sn_variableDelay);

		// Control Memory of Permutation Networks
		this->nodes_permutation_MEM = new RomMemory2<ROM_SIZE, ROM_PERMUTATION_WIDTH, ROM_ADDRESS_WIDTH>("RomPi0");
		this->nodes_permutation_MEM->BindCommonPorts(this);
		this->nodes_permutation_MEM->readEnable(this->nodes_permutation_MEM_enable);
		this->nodes_permutation_MEM->readAddress(this->permutation_index_delayed);
		this->nodes_permutation_MEM->output(this->permutationIndices[0]);
		this->banks_permutation_MEM = new RomMemory2<ROM_SIZE, ROM_PERMUTATION_WIDTH, ROM_ADDRESS_WIDTH>("RomPi2");
		this->banks_permutation_MEM->BindCommonPorts(this);
		this->banks_permutation_MEM->readEnable(this->banks_permutation_MEM_enable);
		this->banks_permutation_MEM->readAddress(this->permutation_index);
		this->banks_permutation_MEM->output(this->permutationIndices_temp[0]);

		// Delays of the Permutation Indices
		this->shift5 = new ShiftRegister<sc_lv<ROM_ADDRESS_WIDTH> >("Shift5", DELAY_PERMUTATION_2 - 1, ROM_ADDRESS_WIDTH);
		this->shift5->input(this->in_addressRom);
		this->shift5->output(this->permutation_index);
		this->shift5->BindCommonPorts(this);
		this->shift11 = new ShiftRegister<sc_logic>("Shift11", DELAY_PERMUTATION_2 - 1, 1);
		this->shift11->input(this->in_enableRom);
		this->shift11->output(this->banks_permutation_MEM_enable);
		this->shift11->BindCommonPorts(this);
		this->shift8 = new ShiftRegister<sc_lv<DEGREE_WIDTH> >("Shift8", DELAY_PERMUTATION_2, DEGREE_WIDTH);
		this->shift8->input(this->sn_variableDelay);
		this->shift8->output(this->pi3_variableDelay);
		this->shift8->BindCommonPorts(this);

		this->shift6 = new ShiftRegister<sc_lv<ROM_ADDRESS_WIDTH> >("Shift6", DURATION_PERMUTATION + 1, ROM_ADDRESS_WIDTH);
		this->shift6->input(this->permutation_index);
		this->shift6->output(this->permutation_index_delayed);
		this->shift6->BindCommonPorts(this);
		this->shift10 = new ShiftRegister<sc_logic>("Shift10", DURATION_PERMUTATION + 1, 1);
		this->shift10->input(this->banks_permutation_MEM_enable);
		this->shift10->output(this->nodes_permutation_MEM_enable);
		this->shift10->BindCommonPorts(this);
		this->shift9 = new ShiftRegister<sc_lv<DEGREE_WIDTH> >("shift9", DURATION_PERMUTATION, DEGREE_WIDTH);
		this->shift9->input(this->pi3_variableDelay);
		this->shift9->output(this->pi1_variableDelay);
		this->shift9->BindCommonPorts(this);

		this->variableShift_nodes_permutation = new VariableShiftRegister<sc_lv<ROM_PERMUTATION_WIDTH> >("NodesPermutationIndicesShift",  
				MAX_DEGREE, DELAY_PERMUTATION_1 - 1, ROM_PERMUTATION_WIDTH);
		this->variableShift_nodes_permutation->BindCommonPorts(this);
		this->variableShift_nodes_permutation->input(this->permutationIndices[0]);
		this->variableShift_nodes_permutation->output(this->permutationIndices_temp[1]);
		this->variableShift_nodes_permutation->active_size(this->pi1_variableDelay);
		this->variableShift_banks_permutation = new VariableShiftRegister<sc_lv<ROM_PERMUTATION_WIDTH> >("BanksPermutationIndicesShift",
				MAX_DEGREE, DELAY_PERMUTATION_3, ROM_PERMUTATION_WIDTH);
		this->variableShift_banks_permutation->BindCommonPorts(this);
		this->variableShift_banks_permutation->input(this->permutationIndices[2]);
		this->variableShift_banks_permutation->output(this->permutationIndices[3]);
		this->variableShift_banks_permutation->active_size(this->pi3_variableDelay);
	}
	
	void RomSNOutput()
	{
		sc_lv<(ROM_SN_WIDTH - 2) * PARALLELISM> oldSD_address;
		for(int i = 0; i < PARALLELISM; i++)
		{
			int max = (PARALLELISM - i) * (ROM_SN_WIDTH - 2) - 1;
			int min = (PARALLELISM - i - 1) * (ROM_SN_WIDTH - 2);
			oldSD_address(max, min) = this->romSN_output.read().range((PARALLELISM - i)*ROM_SN_WIDTH - 1,(PARALLELISM - i - 1)*ROM_SN_WIDTH + 2);
		}
		this->out_oldSD_address_vector.write(oldSD_address);
	}

	void PermutationOutput()
	{
		sc_lv<4 * ROM_PERMUTATION_WIDTH> out_permutationIndices_vector;
		for(int i = 0; i < 4; i++)
		{
			int max = (4 - i) * ROM_PERMUTATION_WIDTH - 1;
			int min = (4 - i - 1) * ROM_PERMUTATION_WIDTH;
			out_permutationIndices_vector(max, min) = this->permutationIndices[i].read();
		}
		this->out_permutationIndices_vector.write(out_permutationIndices_vector);
	}
	
	void Update()
	{
		// Invert indices
		sc_lv<ROM_PERMUTATION_WIDTH> index = permutationIndices_temp[1].read();
		sc_lv<ROM_PERMUTATION_WIDTH> reversed_index = 0;
		if(index.to_uint() != 0)
		{
			for(int i = PARALLELISM; i > 0; i--)
			{
				int temp = index.range(i*PARALLELISM_WIDTH - 1, (i - 1)*PARALLELISM_WIDTH).to_uint();
				reversed_index.range(ROM_PERMUTATION_WIDTH - 1 - temp * PARALLELISM_WIDTH,ROM_PERMUTATION_WIDTH - (temp + 1) * PARALLELISM_WIDTH) = PARALLELISM - i;
			}
			this->permutationIndices[1].write(reversed_index);
		}
		else
		{
			this->permutationIndices[1].write(index);
		}


		if(in_oddIteration.read() == SC_LOGIC_1)
		{
			index = permutationIndices_temp[0].read();
			if(index.to_uint() != 0)
			{
				for(int i = PARALLELISM; i > 0; i--)
				{
					int temp = index.range(i*PARALLELISM_WIDTH - 1, (i - 1)*PARALLELISM_WIDTH).to_uint();
					reversed_index.range(ROM_PERMUTATION_WIDTH - 1 - temp * PARALLELISM_WIDTH,ROM_PERMUTATION_WIDTH - (temp + 1) * PARALLELISM_WIDTH) = PARALLELISM - i;
				}
				this->permutationIndices[2].write(reversed_index);
			}
			else
			{
				this->permutationIndices[2].write(index);
			}
		}
		else
		{
			this->permutationIndices[2].write(this->permutationIndices_temp[0].read());
		}
	}

	void Reset()
	{
		this->permutationIndices[1].write(0);
		this->permutationIndices[2].write(0);
	}
};
#else
class ControlSNandPI : public VHDL_Component
{
public:
	sc_in<sc_logic>										in_oddIteration;
	sc_in<sc_logic>										in_enableRom;
	sc_in<sc_lv<ROM_ADDRESS_WIDTH> >					in_addressRom;
	sc_in<sc_lv<DEGREE_WIDTH> >							in_variableDelay;

	sc_out<sc_logic>									out_newSD_readEnable;
	sc_out<sc_logic>									out_oldSD_readEnable;
	sc_out<sc_lv<(ROM_SN_WIDTH - 2) * PARALLELISM> >	out_oldSD_address_vector;
	sc_out<sc_lv<ROM_SN_WIDTH * PARALLELISM> >			out_newSD_address_vector;
	sc_out<sc_lv<4 * ROM_PERMUTATION_WIDTH> >			out_permutationIndices_vector;

	ControlSNandPI(sc_module_name name) : VHDL_Component(name),
		in_oddIteration("in_oddIteration"),
		in_enableRom("in_enableRom"),
		in_addressRom("in_addressRom"),
		in_variableDelay("in_variableDelay"),
		out_newSD_readEnable("out_newSD_readEnable"),
		out_oldSD_readEnable("out_oldSD_readEnable"),
		out_oldSD_address_vector("out_oldSD_address_vector"),
		out_newSD_address_vector("out_newSD_address_vector"),
		out_permutationIndices_vector("out_permutationIndices_vector")
	{
		elaborate_foreign_module("work.ControlSNAndPi");
    }
};
#endif // MTI_SYSTEMC

#endif // ACCESSROM_H