#ifndef DECODESCHEDULING_H
#define DECODESCHEDULING_H
#include "..\Component.h"
#include "..\Components\VariableShiftRegister.h"

#ifndef MTI_SYSTEMC
class DecodeScheduling : public Component
{
public:

	// interface
	sc_out<sc_logic> out_last_iteration_reached;
	sc_out<sc_lv<ROM_ADDRESS_WIDTH> > out_rom_index;
	sc_out<sc_logic> out_rom_enable;
	sc_out<sc_logic> out_first_value;						//cn
	sc_out<sc_lv<DEGREE_WIDTH> > out_cn_index;				//cn
	sc_out<sc_lv<DEGREE_WIDTH> > out_cn_degree;				//cn
	sc_out<sc_lv<WIDTH_ITERATION> > out_iteration;

	sc_signal<sc_logic> last_iteration_reached;

	// memories
	sc_signal<int>* MEM_degree;
	sc_signal<int>* MEM_degree_number;
	sc_signal<int> index_MEM_control;
	sc_signal<int> index_MEM_degree;
	sc_signal<bool> enable_MEM_degree;
	sc_signal<bool> enable_MEM_degree_delayed;
	int MEM_last_index;
	
	// Reading of the next degree
	sc_signal<int> new_degree;
	sc_signal<int> index;
	sc_signal<int> count_degree;
	sc_signal<int> number_of_degrees;
	sc_signal<int> iteration;
	sc_signal<bool> enable_next_degree_count;
	sc_signal<bool> all_degrees_read_for_iteration;
	sc_signal<bool> all_degrees_read_for_iteration_delayed;
	sc_signal<int> counter_delay;
	int delay_new_degree_update;
	int iteration_delay_overhead;

	// Generation of Control Signals
	sc_signal<int> current_degree;
	sc_signal<int> counter_output_degree;
	sc_signal<int> cn_read_address;
	sc_signal<int> cn_write_address;
	sc_signal<int> cn_mem_address;
	sc_signal<bool> cn_read_enable;
	sc_signal<bool> cn_read_enable_delayed1;
	sc_signal<bool> cn_read_enable_delayed2;

	sc_signal<bool> point_A;
	sc_signal<int> point_C;
	bool iteration_delay;
	sc_signal<bool> iteration_delay_delayed;
	sc_signal<bool> point_no_delay;

	ShiftRegister<int>* shift1;
	SC_HAS_PROCESS(DecodeScheduling);
	int max_iteration;
	DecodeScheduling(sc_module_name name, const char* control_file, int max_iteration)
	{
		SC_METHOD(Points_A_C);
		sensitive << this->new_degree << this->current_degree;
		SC_METHOD(DelayBetweenDescendingDegreesOrIterations);
		sensitive << this->counter_output_degree << this->point_C << this->point_A << this->counter_delay;
		SC_METHOD(EndOfIteration);
		sensitive << this->index_MEM_degree << this->enable_MEM_degree;
		this->ReadControlInformation(control_file);
		this->delay_new_degree_update = 3;
		this->iteration_delay_overhead = DELAY_NEW_SD + this->delay_new_degree_update + 20;
		this->max_iteration = max_iteration;
	}

	void ReadControlInformation(const char* fileName)
	{
		FILE *file_Control = NULL;
		int temp;

		// open file to read
#ifdef MTI_SYSTEMC
		file_Control = fopen(fileName,"r");
#else
		if (fopen_s(&file_Control, fileName,"r") != NULL) 
		{ 
			printf("Failed to open file containing control information.\n"); 
		}
#endif //MTI_SYSTEMC

		int degrees;
		temp = fseek(file_Control, 0, SEEK_SET);
		temp = fscanf(file_Control,"%i", &degrees);

		this->MEM_degree = new sc_signal<int>[degrees];
		for(int i = 0; i < degrees; i++)
		{
			fscanf(file_Control,"%i", &temp);
			this->MEM_degree[i].write(temp);
		}
		this->MEM_degree_number = new sc_signal<int>[degrees];
		for(int i = 0; i < degrees; i++)
		{
			fscanf(file_Control,"%i", &temp);
			this->MEM_degree_number[i].write(temp);
		}
		
		this->MEM_last_index = degrees - 1;
		fclose(file_Control);
	}

	void DelayBetweenDescendingDegreesOrIterations()
	{	
		bool delay_between_descending_degrees = !this->point_A.read() || (this->counter_output_degree.read() == this->point_C.read());
		this->iteration_delay = this->counter_delay.read() > 0 ? true : false;
		this->point_no_delay.write(delay_between_descending_degrees && !(this->iteration_delay || this->all_degrees_read_for_iteration_delayed.read()));
	}

	void EndOfIteration()
	{
		int index_MEM_degree = this->index_MEM_degree.read();
		bool last_degree_reached = (this->iteration.read() % 2) == 0 ? index_MEM_degree == this->MEM_last_index : index_MEM_degree == 0;
		bool enable_MEM_degree = this->enable_MEM_degree.read();
		this->enable_next_degree_count.write(enable_MEM_degree && !last_degree_reached);
		this->all_degrees_read_for_iteration.write(enable_MEM_degree && last_degree_reached);
	}

	void Points_A_C()
	{
		int current_degree = this->current_degree.read();
		int new_degree = this->new_degree.read();
		bool point_A = (current_degree > new_degree) ? true : false;
		int point_D = current_degree - new_degree;
		int point_B = current_degree + point_D;
		this->point_C.write((point_A == true) ? point_B : current_degree);
		this->point_A.write(point_A);
	}

	void Update()
	{
		this->out_last_iteration_reached.write(this->last_iteration_reached.read());
		if(this->last_iteration_reached.read() == SC_LOGIC_0)
		{
			/*out_rom_index = out_rom_index;
			out_rom_enable = out_rom_enable;
			out_first_value = out_first_value;						//cn
			out_cn_index = out_cn_index;				//cn
			out_cn_degree = out_cn_degree;				//cn
			out_iteration = out_iteration;
			*/
				// Shift Registers
			this->iteration_delay_delayed.write(this->iteration_delay);
			this->all_degrees_read_for_iteration_delayed.write(this->all_degrees_read_for_iteration.read());

			this->out_first_value.write(SC_LOGIC_0);
			this->enable_MEM_degree.write(false);
			this->out_rom_enable.write(SC_LOGIC_0);
			int temp;
			// --------------------------- Read Memory Addresses -------------------------- //
			temp = this->index_MEM_control.read();
			if(this->point_no_delay.read() == true)
			{
				if((this->iteration.read() % 2) == 0)
				{
					temp++;
				}
				else
				{
					temp--;
				}
			}
			this->index_MEM_control.write(temp);
			this->out_rom_index.write((unsigned int)this->index_MEM_control.read());
			this->out_cn_index.write(this->index.read());

			// ----------------------- Generation of Control Signals ---------------------- //
			// counter degree
			temp = this->counter_output_degree.read();
			if(temp == point_C)
			{
				temp = 0;
			}
			else if(this->iteration_delay_delayed.read() == false)
			{
				temp++;
			}
			this->counter_output_degree.write(temp);

			if(this->iteration_delay == false)
			{
				if(temp == 0)
				{
					this->out_first_value.write(SC_LOGIC_1);
				}
				if(temp <= this->current_degree.read())
				{
					if(this->iteration.read() <= this->max_iteration - 1)
					{
						this->out_rom_enable.write(SC_LOGIC_1);
					}
					else
					{
						this->out_rom_enable.write(SC_LOGIC_0);
					}
				}
			}


			// current degree
			/*if(this->iteration_delay == true)
			{
				temp = 0;
			}
			else*/ if(this->point_no_delay.read() == true)
			{
				temp = this->new_degree.read();
			}
			else
			{
				temp = this->current_degree.read();
			}
			this->current_degree.write(temp);
			this->out_cn_degree.write(temp);

			// ------------------------ Reading of the next degree ------------------------ //

			// memory accesses
			if(this->enable_MEM_degree_delayed.read() == true)
			{
				temp = index_MEM_degree.read();
				this->new_degree.write(this->MEM_degree[temp].read());
				this->number_of_degrees.write(this->MEM_degree_number[temp].read());
			}

			// index counter
			temp = this->index.read();
			if(temp == this->new_degree.read() || this->all_degrees_read_for_iteration_delayed.read() == true)
			{
				temp = 0;
			}
			else if(this->point_no_delay.read() == true)
			{
				temp++;
			}
			this->index.write(temp);
		
			// count degree counter
			temp = this->count_degree.read();
			bool count_degree_enable = (this->point_no_delay.read() == true) && (this->index.read() == (this->new_degree.read() - this->delay_new_degree_update));
			if(temp == this->number_of_degrees.read())
			{
				this->enable_MEM_degree.write(true);
				temp = 0;
			}
			else if(this->all_degrees_read_for_iteration_delayed.read() == true)
			{
				temp = 0;
			}
			else if(count_degree_enable)
			{
				temp++;
			}
			this->count_degree.write(temp);

			// CN MEM Address for read and write
			this->cn_read_enable.write(count_degree_enable);
			this->cn_read_enable_delayed2.write(this->cn_read_enable_delayed1.read());
			this->cn_read_enable_delayed1.write(this->cn_read_enable.read());
			if(this->cn_read_enable_delayed1.read() == true && !this->all_degrees_read_for_iteration == true)
			{
				temp = this->cn_read_address.read();
				if((this->iteration.read() % 2) == 0)
				{
					temp++;
				}
				else
				{
					temp--;
				}
				this->cn_read_address.write(temp);
			
			}
			if(this->cn_read_enable_delayed1.read() == true)
			{
				this->cn_write_address.write(this->cn_read_address.read());
			}
			this->cn_mem_address.write((this->cn_read_enable_delayed2.read() == true || this->counter_output_degree.read() >= this->current_degree.read() ) ? this->cn_read_address.read() : this->cn_write_address.read());
			//this->out_cn_address.write(this->cn_mem_address.read());
		
			// index of degree counter
			temp = this->index_MEM_degree.read();
			if(this->enable_next_degree_count.read())
			{
				if((this->iteration.read() % 2) == 0)
				{
					temp++;
				}
				else
				{
					temp--;
				}
			}
			this->index_MEM_degree.write(temp);

			// iteration counter
			temp = this->iteration.read();
			if(this->all_degrees_read_for_iteration_delayed.read())
			{
				temp++;
			}
			this->iteration.write(temp);

			this->iteration_delay = this->counter_delay.read() > 0 ? true : false;
			if(!this->iteration_delay)
			{
				this->out_iteration.write(this->iteration.read());
			}
			if(this->counter_delay.read() == 5)
			{
				if(this->iteration.read() == this->max_iteration)
				{
					this->out_iteration.write(this->iteration.read());
					this->out_last_iteration_reached.write(SC_LOGIC_1);
				}
			}

			// enable Memory access
			if(this->enable_next_degree_count.read())
			{
				this->enable_MEM_degree_delayed.write(this->enable_MEM_degree.read());
			}
		
			// counter delay
			if(this->all_degrees_read_for_iteration_delayed.read())
			{
				this->counter_delay.write(this->new_degree.read() + this->iteration_delay_overhead);		// initialization
			}
			temp = this->counter_delay.read();
			if(temp > 0)
			{
				temp--;
				this->counter_delay.write(temp);
			}
		}
	}

	void Reset()
	{
		this->new_degree.write(this->MEM_degree[0].read());
		this->number_of_degrees.write(this->MEM_degree_number[0].read());
		this->index.write(0);
		this->count_degree.write(0);
		this->index_MEM_degree.write(0);
		this->counter_delay.write(0);
		this->iteration.write(0);
		this->index_MEM_control.write(0);
		this->current_degree.write(0);
		this->counter_output_degree.write(0);
		this->counter_delay.write(0);
		this->all_degrees_read_for_iteration_delayed.write(false);
		this->iteration_delay_delayed.write(false);
		this->out_iteration.write(0);
		this->out_cn_degree.write(this->MEM_degree[0].read());
		this->out_rom_index.write(0);
		this->cn_read_address.write(0);
		this->cn_write_address.write(0);
		this->cn_read_enable.write(0);
		this->out_rom_enable.write(SC_LOGIC_0);
		//this->out_cn_address.write(0);
		this->out_cn_index.write(0);
		this->out_last_iteration_reached.write(SC_LOGIC_0);
		this->last_iteration_reached.write(SC_LOGIC_0);
	}
};
#else
class DecodeScheduling : public VHDL_Component
{
public:
	sc_out<sc_lv<ROM_ADDRESS_WIDTH> > out_rom_index;
	sc_out<sc_logic> out_rom_enable;
	sc_out<sc_logic> out_first_value;						
	sc_out<sc_lv<DEGREE_WIDTH> > out_cn_index;				
	sc_out<sc_lv<DEGREE_WIDTH> > out_cn_degree;				
	sc_out<sc_lv<WIDTH_ITERATION> > out_iteration;
	sc_out<sc_logic> out_last_iteration_reached;

    DecodeScheduling(sc_module_name name, const char* control_file, int max_iteration) : VHDL_Component(name),
		out_rom_index("out_rom_index"),
		out_rom_enable("out_rom_enable"),
		out_first_value("out_first_value"),					
		out_cn_index("out_cn_index"),			
		out_cn_degree("out_cn_degree"),		
		out_iteration("out_iteration"),
		out_last_iteration_reached("out_lastIterationReached")
    {
		this->add_parameter("widthIteration", WIDTH_ITERATION);
		this->add_parameter("maxIteration", MAX_ITERATION);
		elaborate_foreign_module("work.DecodeScheduling");
    }
};
#endif // MTI_SYSTEMC
#endif //DECODESCHEDULING_H
