#ifndef FINITESTATEMACHINE_H
#define FINITESTATEMACHINE_H
#include "..\Component.h"
#ifndef MTI_SYSTEMC
enum state{IDLE, INIT, DECODE, WRITEBACK, INIT_WRITEBACK};

class FiniteStateMachine : public Component
{
public:
	sc_in<sc_logic>					in_new_values_present;					// this flag determines wether the next received code word is available
	sc_out<sc_logic>				out_ready_for_next_received_value;		// this flag indicates, that the decoder ist ready for the next received symbol
	sc_in<sc_logic>					in_decoding_valid;						// indicates success of the decoding
	sc_out<sc_logic>				out_decoding_valid;
	sc_in<sc_lv<WIDTH_ITERATION> >	in_current_iteration;					// current iteration of the decoding process
	sc_out<sc_logic>				out_enable_decode_schedulig;			// if true the decoding starts
	sc_out<sc_logic>				out_reset_decode_scheduling;
	sc_out<sc_lv<WIDTH_ITERATION> >	out_iterations;
	sc_in<sc_logic>					in_last_iteration_reached;
	sc_out<sc_logic>				out_init;
	sc_out<sc_logic>				out_writeBack;
	sc_in<sc_logic>					in_loadAndWriteBackDone;

	sc_signal<state> current_state;

	FiniteStateMachine(sc_module_name name)
	{
	}

	void Update()
	{
		// preload signals
		state next_state = this->current_state.read();
		this->out_ready_for_next_received_value.write(SC_LOGIC_0);
		this->out_enable_decode_schedulig.write(SC_LOGIC_0);
		this->out_reset_decode_scheduling.write(SC_LOGIC_0);
		this->out_init.write(SC_LOGIC_0);
		this->out_writeBack.write(SC_LOGIC_0);

		// FSM
		switch(this->current_state.read())
		{
		case IDLE:
			if(this->in_new_values_present.read() == SC_LOGIC_1)
			{
				this->out_ready_for_next_received_value.write(SC_LOGIC_1);
				next_state = INIT;
			}
			break;
		case INIT:
			{
				this->out_init.write(SC_LOGIC_1);
				if(this->in_loadAndWriteBackDone == SC_LOGIC_1)
				{
					next_state = DECODE;
				}
			}
			break;
		case DECODE:
			{
				int iteration = this->in_current_iteration.read().to_uint();
				if((in_last_iteration_reached.read() | this->in_decoding_valid.read())  == SC_LOGIC_1)
				{
					this->out_reset_decode_scheduling.write(SC_LOGIC_1);
					this->out_decoding_valid.write(this->in_decoding_valid.read());
					this->out_iterations.write(iteration);
					if(this->in_new_values_present.read() == SC_LOGIC_1)
					{
						next_state = INIT_WRITEBACK;
						this->out_ready_for_next_received_value.write(SC_LOGIC_1);
					}
					else
					{
						next_state = WRITEBACK;
					}
				}
				else
				{
					this->out_enable_decode_schedulig.write(SC_LOGIC_1);
				}
			}
			break;
		case WRITEBACK:
			{
				this->out_writeBack.write(SC_LOGIC_1);
				if(this->in_loadAndWriteBackDone == SC_LOGIC_1)
				{
					next_state = IDLE;
				}
			}
			break;
		case INIT_WRITEBACK:
			{
				this->out_init.write(SC_LOGIC_1);
				this->out_writeBack.write(SC_LOGIC_1);
				if(this->in_loadAndWriteBackDone == SC_LOGIC_1)
				{
					next_state = DECODE;
				}
			}
			break;
		default:
			break;
		}
		// write next state
		this->current_state.write(next_state);
	}

	void Reset()
	{
		this->current_state.write(IDLE);
		this->out_reset_decode_scheduling.write(SC_LOGIC_1);
	}
};
#else
class FiniteStateMachine : public VHDL_Component
{
public:
	sc_in<sc_logic>					in_new_values_present;					// this flag determines wether the next received code word is available
	sc_out<sc_logic>				out_ready_for_next_received_value;		// this flag indicates, that the decoder ist ready for the next received symbol
	sc_in<sc_logic>					in_decoding_valid;						// indicates success of the decoding
	sc_out<sc_logic>				out_decoding_valid;
	sc_in<sc_lv<WIDTH_ITERATION> >	in_current_iteration;					// current iteration of the decoding process
	sc_out<sc_logic>				out_enable_decode_schedulig;			// if true the decoding starts
	sc_out<sc_logic>				out_reset_decode_scheduling;
	sc_out<sc_lv<WIDTH_ITERATION> >	out_iterations;
	sc_in<sc_logic>					in_last_iteration_reached;
	sc_out<sc_logic>				out_init;
	sc_out<sc_logic>				out_writeBack;
	sc_in<sc_logic>					in_loadAndWriteBackDone;

	FiniteStateMachine(sc_module_name name) : VHDL_Component(name),
		in_new_values_present("in_newValuesPresent"),					// this flag determines wether the next received code word is available
		out_ready_for_next_received_value("out_readyForNextCodeword"),		// this flag indicates, that the decoder ist ready for the next received symbol
		in_decoding_valid("in_decodingValid"),						// indicates success of the decoding
		out_decoding_valid("out_decodingValid"),
		in_current_iteration("in_currentIteration"),					// current iteration of the decoding process
		out_enable_decode_schedulig("out_enableDecodeSchedulig"),			// if true the decoding starts
		out_reset_decode_scheduling("out_resetDecodeScheduling"),
		out_iterations("out_iterations"),
		in_last_iteration_reached("in_lastIterationDone"),
		out_init("out_initState"),
		out_writeBack("out_outputState"),
		in_loadAndWriteBackDone("in_initAndOutputDone")
	{
		this->add_parameter("WIDTH_ITERATION", WIDTH_ITERATION);
		elaborate_foreign_module("work.FiniteStateMachine");
	}
};
#endif // MTI_SYSTEMC

#endif // FINITESTATEMACHINE_H