#ifndef SHIFT_REGISTER_MONITOR_H
#define SHIFT_REGISTER_MONITOR_H
#include "Monitor.h"
#include "..\Decoder\Components\VariableShiftRegister.h"

class ShiftRegisterMonitor : public Monitor
{
public:
	VariableShiftRegister<sc_lv<10> >* shiftRegister;
	sc_signal<sc_lv<10> > input;
	sc_signal<sc_lv<10> > output;
	sc_signal<sc_lv<3> > activeSize;
	int size;
	int* values;

	ShiftRegisterMonitor(sc_module_name name) : Monitor(name)
	{
		this->size = 5;
		int delay = this->size + 1;
		this->shiftRegister = new VariableShiftRegister<sc_lv<10> >("shiftRegister", delay, 0, 10);
		this->shiftRegister->active_size(this->activeSize);

		this->shiftRegister->clock(this->iclock);
		this->shiftRegister->enable(this->enable);
		this->shiftRegister->reset(this->reset);
		this->shiftRegister->input(this->input);
		this->shiftRegister->output(this->output);
		this->activeSize = 4;
		
		int values[24] = {2, -7, -3, -1, 9, -8, 3, 5, -2, 1, 4, 6, 2, -4, 3, 3, -4, 4, -6, 3, 6, -1, 3, 7};
		this->values = new int[24];
		for(int i = 0; i < 24; ++i)
		{
			this->values[i] = values[i];
		}
		/*bool values[24] = {1, 1, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 1};
		this->values = new bool[24];
		for(int i = 0; i < 24; ++i)
		{
			this->values[i] = values[i];
		}*/
	}

	void TraceSignals(sc_trace_file *tf)
	{
#ifndef MTI_SYSTEMC
		this->shiftRegister->TraceSignals(tf);
#endif
	}

	void CyclicUpdate()
	{
		int i = 0;
		while(1)
		{
			this->input.write(this->values[i]);
			i++;
			if(i == 23)
			{
				break;
			}
			wait();
		}
		wait(200);
		sc_stop();
	}

	void SequentialUpdate()
	{
		wait(5);
		this->activeSize.write(4);
		wait(10);
		this->activeSize.write(3);
		wait(5);
		this->activeSize.write(0);
		wait(5);
	}
};
#endif //SHIFT_REGISTER_MONITOR_H
