-- Filename     : ROM.vhd
-- Author:	    : Andreas Torno
-- Description  : ROM memory wich is initialized from mif file

library ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;
USE std.textio.all;
USE ieee.std_logic_textio.all;

ENTITY ROM IS
	GENERIC (
		size 				: integer := 1;
		width_data 			: integer := 32;
		width_address		: integer := 32;
		filename			: string := "rom"
	);
	
    PORT (
		-- Control ports
		clock		: IN std_logic;
		enable		: IN std_logic;
		reset		: IN std_logic;
		r_enable	: IN std_logic;
		address		: IN std_logic_vector(width_address - 1 downto 0);
		
		-- Data ports
        output    	: OUT std_logic_vector(width_data - 1 downto 0)
    );
END ROM;

ARCHITECTURE RTL OF ROM IS
	type memory is array(0 to size - 1) of std_logic_vector(width_data - 1 downto 0);
	
	impure function InitRomFromFile (RamFileName : in string) return memory is
		FILE 		RamFile 	: text is in RamFileName;
		variable 	RamFileLine : line;
		variable 	RAM 		: memory;
	begin
		for I in memory'range loop
			readline (RamFile, RamFileLine);
			read (RamFileLine, RAM(I));
		end loop;
		return RAM;
	end function;
	
	CONSTANT ROM_LOCATION					: string  := "D:\DecoderWorkspace\ControlInformation\";
	signal memory_content 					: memory := InitRomFromFile(ROM_LOCATION & filename & ".mif");
	attribute RAM_STYLE 					: string;
	attribute RAM_STYLE of memory_content 	: signal is "BLOCK"; --{ "DISTRIBUTED" | "BLOCK" };
	signal output_temp    					: std_logic_vector(width_data - 1 downto 0);
BEGIN
	process(clock)
    begin
	if clock'event and clock='1' then
		if reset = '1' then
			output <= (others => '0');
		elsif enable = '1' then
			if r_enable = '1' then
				output <= output_temp;
			end if;
		end if;
	end if;
  end process;
  output_temp	<= memory_content(to_integer(unsigned(address)));
END RTL;
