-- Filename     : MemorySinglePort.vhd
-- Author:	    : Andreas Torno
-- Description  : Although this memory has 2 data ports, it is considered as single port memory
--				: because it has only one address port. Thus its physical realisation is smaller
--				: as of dual port memory which has twice as much address wires.

library ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;

ENTITY MemorySinglePort IS
	GENERIC (
		size 				: integer := 204;
		width_data 			: integer := 8;
		width_address		: integer := 8
	);
	
    PORT (
		-- Control ports
		clock		: IN std_logic;
		enable		: IN std_logic;
		reset		: IN std_logic;
		rw_enable	: IN std_logic;
		address		: IN std_logic_vector(width_address - 1 downto 0);
		
		-- Data ports
        input     	: IN std_logic_vector(width_data - 1 downto 0);
        output    	: OUT std_logic_vector(width_data - 1 downto 0)
    );
END MemorySinglePort;

ARCHITECTURE RTL OF MemorySinglePort IS
	type memory is array(size - 1 downto 0) of std_logic_vector(width_data - 1 downto 0);
	
	impure function Initialize return memory is
		variable RAM : memory;
	begin
		for I in memory'range loop
			RAM(i) := (OTHERS => '0');
		end loop;
		return RAM;
	end function;
	
	signal memory_content : memory := Initialize;
	attribute RAM_STYLE : string;
	attribute RAM_STYLE of memory_content : signal is "DISTRIBUTED"; --{ "DISTRIBUTED" | "BLOCK" };
BEGIN
	process(clock)
    begin
	-- if (rising_edge(clock)) then
		-- if reset = '1' then
			-- output <= (others => '0');
	    -- elsif enable = '1' then
			-- if rw_enable = '0' then
				-- memory_content(to_integer(unsigned(address))) <= input;
			-- else
				-- output <= memory_content(to_integer(unsigned(address)));
			-- end if;
	    -- end if;
	-- end if;
	if (rising_edge(clock)) then
		if reset = '1' then
			output <= (others => '0');
		elsif enable = '1' then
			if rw_enable = '0' then
				memory_content(to_integer(unsigned(address))) <= input;
			elsif rw_enable = '1' then
				output <= memory_content(to_integer(unsigned(address)));
			end if;
		end if;
	end if;
    end process;
END RTL;
