--
-- Packages containing different data types
-- by Andreas Torno
-- 

LIBRARY ieee;
USE ieee.std_logic_1164.all;
use ieee.numeric_std.all;

PACKAGE types IS
	CONSTANT WIDTH_SD : integer := 8;
	CONSTANT WIDTH_MAGNITUDE : integer := WIDTH_SD - 1;
	TYPE SD_TYPE IS ARRAY (WIDTH_SD - 1 DOWNTO 0) OF std_logic;
	--TYPE MAGNITUDE_TYPE IS ARRAY (WIDTH_MAGNITUDE - 1 DOWNTO 0) OF std_ulogic;
	TYPE SD_TYPE_vector IS ARRAY ( NATURAL RANGE <> ) OF std_logic_vector(WIDTH_SD - 1 downto 0);
	
	FUNCTION "+" (op1, op2 : in SD_TYPE) return SD_TYPE;
	FUNCTION "-" (op1, op2 : in SD_TYPE) return SD_TYPE;
	FUNCTION "-" (op : in SD_TYPE) return SD_TYPE;
	--FUNCTION abs_min (op1, op2 : in SD_TYPE) return MAGNITUDE_TYPE;
END types;

PACKAGE BODY types IS
	--  435 (magnitude_of_input_message -> min2) - SN 415 (in_SDwrite -> sd1_input or intrinsic_Memory.output -> out_SD_write)
	FUNCTION "+" (op1, op2 : in SD_TYPE) return SD_TYPE IS
	VARIABLE sum : std_logic_vector (WIDTH_SD - 1 DOWNTO 0);
	VARIABLE output : std_logic_vector (WIDTH_SD - 1 DOWNTO 0);
	CONSTANT possibleMin : std_logic_vector (WIDTH_SD - 1 DOWNTO 0) := (WIDTH_SD - 1 => '1', OTHERS => '0');
	BEGIN
		sum := std_logic_vector(unsigned(op1) + unsigned(op2));
		IF (op1(WIDTH_SD - 1) = op2(WIDTH_SD - 1) and sum(WIDTH_SD - 1) /= op1(WIDTH_SD - 1)) THEN
			output := (OTHERS => not op1(WIDTH_SD - 1));
			output(WIDTH_SD - 1) := op1(WIDTH_SD - 1);
			IF op1(WIDTH_SD - 1) = '1' THEN
				output(0) := '1';
			END IF;
			return SD_TYPE(output);
		ELSIF sum = possibleMin then
			output := sum;
			output(0) := '1';
			return SD_TYPE(output);
		end if;
		return SD_TYPE(sum);
	END FUNCTION "+";
	
	-- CN 431 - SN 400
	-- FUNCTION "+" (op1, op2 : in SD_TYPE) return SD_TYPE IS
	-- VARIABLE temp : std_ulogic_vector ( 0 TO WIDTH_SD - 1);
	-- CONSTANT possibleMin : std_ulogic_vector ( 0 TO WIDTH_SD - 1) := (0 to 0 => '1', OTHERS => '0');
	-- CONSTANT min : std_ulogic_vector ( 0 TO WIDTH_SD - 1) := (0 to 0 => '1', 1 to WIDTH_SD - 2 => '0', WIDTH_SD - 1 => '1');
	-- CONSTANT max : std_ulogic_vector ( 0 TO WIDTH_SD - 1) := (0 to 0 => '0', OTHERS => '1');
	-- BEGIN
		-- temp := std_ulogic_vector(unsigned(op1) + unsigned(op2));
		-- IF (op1(0) = op2(0) and temp(0) /= op1(0)) THEN
			-- IF op1(0) = '1' THEN
				-- return SD_TYPE(min);
			-- ELSE
				-- return SD_TYPE(max);
			-- END IF;
		-- ELSIF temp = possibleMin then
			-- return SD_TYPE(min);
		-- ELSE
			-- return SD_TYPE(temp);
		-- end if;
	-- END FUNCTION "+";
	
	-- CN 435 - SN 401 -- wrong simulation results
	-- FUNCTION "+" (op1, op2 : in SD_TYPE) return SD_TYPE IS  
	-- VARIABLE temp : std_ulogic_vector ( 0 TO WIDTH_SD);
	-- CONSTANT min : std_ulogic_vector ( 0 TO WIDTH_SD - 1) := (0 to 0 => '1', 1 to WIDTH_SD - 2 => '0', WIDTH_SD - 1 => '1');
	-- CONSTANT max : std_ulogic_vector ( 0 TO WIDTH_SD - 1) := (0 to 0 => '0', OTHERS => '1');
	-- BEGIN
		-- temp := std_ulogic_vector(('0' & unsigned(op1)) + ('0' & unsigned(op2)));
		-- IF signed(temp) < signed(min) THEN
			-- return SD_TYPE(min);
		-- ELSIF signed(temp) > signed(max) THEN
			-- return SD_TYPE(max);
		-- ELSE
			-- return SD_TYPE(temp(0 TO WIDTH_SD - 1));
		-- END IF;
	-- END FUNCTION "+";
	
	FUNCTION "-" (op1, op2 : in SD_TYPE) return SD_TYPE IS
	BEGIN
		return op1 + (-op2);
	END FUNCTION "-";
	
	FUNCTION "-" (op : in SD_TYPE) return SD_TYPE IS
	CONSTANT temp : std_ulogic_vector ( 0 TO WIDTH_SD - 1) := (OTHERS => '1');
	BEGIN
		return SD_TYPE(std_ulogic_vector(unsigned(temp) - unsigned(op) + 1));
	END FUNCTION "-";
	
	-- FUNCTION abs_min (op1, op2 : in MAGNITUDE_TYPE) return MAGNITUDE_TYPE IS
	-- variable temp1, temp2 : integer;
	-- BEGIN
		-- IF  abs(signed(op1)) < abs(signed(op2)) THEN
			-- return MAGNITUDE_TYPE(abs(signed(op1)));
		-- ELSE
			-- return MAGNITUDE_TYPE(abs(signed(op2)));
		-- END IF;
	-- END FUNCTION abs_min;
END types;
