-- Filename    : PermutationNetwork.vhd
-- Author:	   : Andreas Torno
-- Description : This Module describes a permutation network letting synthesis tool deside wich implementation is the appropriate

library ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;
library work;
USE work.types.all;
USE work.widths.all;

ENTITY PermutationNetwork IS
    PORT (
		-- Control ports
		clock		: IN  std_logic;
		enable		: IN  std_logic;
		reset		: IN  std_logic;
		
		-- Data ports
        in_input	: IN  std_logic_vector(PARALLELISM * WIDTH_SD - 1 DOWNTO 0);
		in_index	: IN  std_logic_vector(ROM_PERMUTATION_WIDTH - 1 downto 0);
        out_output  : OUT std_logic_vector(PARALLELISM * WIDTH_SD - 1 DOWNTO 0)
    );

END PermutationNetwork;

-- ENTITY PermutationNetwork IS
    -- PORT (
		-- Control ports
		-- clock		: IN  std_logic;
		-- enable		: IN  std_logic;
		-- reset		: IN  std_logic;
		
		-- Data ports
        -- in_input	: IN  SD_TYPE_vector(0 to PARALLELISM - 1);
		-- in_index	: IN  std_logic_vector(ROM_PERMUTATION_WIDTH - 1 downto 0);
        -- out_output  : OUT SD_TYPE_vector(0 to PARALLELISM - 1)
    -- );

-- END PermutationNetwork;

-- ARCHITECTURE RTL OF PermutationNetwork IS

-- BEGIN
	-- permutation: process(clock)
	-- variable index 		: integer range 0 to PARALLELISM - 1;
    -- begin
	-- if clock'event and clock='1' then
		-- if reset = '1' then
			-- for i in 0 to (PARALLELISM - 1) LOOP
				-- out_output(i)	<=	(OTHERS => '0');
		-- end LOOP;
	    -- elsif enable = '1' then
			-- for i in 0 to (PARALLELISM - 1) LOOP
				-- index := to_integer(unsigned(in_index((PARALLELISM - i) * PARALLELISM_WIDTH - 1 downto (PARALLELISM - i - 1) * PARALLELISM_WIDTH)));
				-- out_output(index)	<= in_input(i);
			-- end LOOP;
	    -- end if;
	-- end if;
    -- end process;
-- END RTL;

ARCHITECTURE RTL OF PermutationNetwork IS -- 1200MHZ at p = 3 (static linking to the arrays in combinatorical functions)
TYPE sd_array is ARRAY (0 to PARALLELISM - 1) of std_logic_vector(WIDTH_SD - 1 DOWNTO 0);
signal input		: sd_array;
signal output		: sd_array;

BEGIN
	inputLinking: process(in_input)
	begin
		for i in 0 to (PARALLELISM - 1) LOOP
			input(i)	<= in_input((PARALLELISM - i) * WIDTH_SD - 1 downto (PARALLELISM - i - 1) * WIDTH_SD);
		end LOOP;
	end process;
	
	outputLinking: process(output)
	begin
		for i in 0 to (PARALLELISM - 1) LOOP
			out_output((PARALLELISM - i) * WIDTH_SD - 1 downto (PARALLELISM - i - 1) * WIDTH_SD) <= output(i);
		end LOOP;
	end process;
	
	permutation: process(clock)
	variable index 		: integer range 0 to PARALLELISM - 1;
    begin
	if clock'event and clock='1' then
		if reset = '1' then
			for i in 0 to (PARALLELISM - 1) LOOP
				output(i)	<=	(OTHERS => '0');
		end LOOP;
	    elsif enable = '1' then
			for i in 0 to (PARALLELISM - 1) LOOP
				index := to_integer(unsigned(in_index((PARALLELISM - i) * PARALLELISM_WIDTH - 1 downto (PARALLELISM - i - 1) * PARALLELISM_WIDTH)));
				output(index)	<= input(i);
			end LOOP;
	    end if;
	end if;
    end process;
END RTL;

-- ARCHITECTURE RTL_slow OF PermutationNetwork IS --600 MHZ at p=3 (uses MACs for max and min indices calculation)
-- BEGIN

	-- process(clock)
	-- variable index 		: integer range 0 to PARALLELISM - 1;
	-- variable max_in 		: integer range 0 to PARALLELISM * WIDTH_SD - 1;
	-- variable min_in 		: integer range 0 to PARALLELISM * WIDTH_SD - 1;
	-- variable max_out 	: integer range 0 to PARALLELISM * WIDTH_SD - 1;
	-- variable min_out 	: integer range 0 to PARALLELISM * WIDTH_SD - 1;
    -- begin
	-- if clock'event and clock='1' then
		-- if reset = '1' then
				-- out_output		<= (OTHERS => '0');
	    -- elsif enable = '1' then
			-- for i in 0 to (PARALLELISM - 1) LOOP
				-- index := to_integer(unsigned(in_index((PARALLELISM - i) * PARALLELISM_WIDTH - 1 downto (PARALLELISM - i - 1) * PARALLELISM_WIDTH)));
				-- max_in			:= (PARALLELISM - i) * WIDTH_SD - 1;
				-- min_in			:= (PARALLELISM - i - 1) * WIDTH_SD;
				-- max_out			:= (PARALLELISM - index) * WIDTH_SD - 1;
				-- min_out			:= (PARALLELISM - index - 1) * WIDTH_SD;
				-- out_output(max_out downto min_out) 	<= in_input(max_in downto min_in);
			-- end LOOP;
	    -- end if;
	-- end if;
    -- end process;
-- END RTL_slow;

