-- Filename    : VariableShiftRegisterCN.vhd
-- Author:	   : Andreas Torno
-- Description : Generic variable shift register for check node (signs)
--			   : with possibility to read content for further saving

library ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;

library work;
USE work.widths.all;

ENTITY VariableShiftRegisterCN IS
	GENERIC (
		size 		: integer := 1
	);
	
    PORT (
		-- Control ports
		clock		: IN  std_logic;
		enable		: IN  std_logic;
		reset		: IN  std_logic;
		active_size : IN  std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		read_enable : IN  std_logic;
		
		-- Data ports
        input     	: IN  std_logic;
        output    	: OUT std_logic;
		content		: OUT std_logic_vector(size - 1 downto 0)
    );
END VariableShiftRegisterCN;

ARCHITECTURE RTL OF VariableShiftRegisterCN IS
	signal shift_register :std_logic_vector(size downto 0);
BEGIN
	process(clock)
    begin
	if clock'event and clock='1' then
		if reset = '1' then
			shift_register <= (OTHERS => '0');
			content <= (OTHERS => '0');
	    elsif enable = '1' then
			for i in 1 to size LOOP
				shift_register(i) <= shift_register(i - 1); 
			end LOOP;
			shift_register(size - 1 - to_integer(unsigned(active_size)) ) <= input;
			if read_enable ='1' then
				for i in 0 to size - 1 loop
					content(i) <= shift_register(size - 1 - i);
				end loop;
			end if;
	    end if;
	end if;
    end process;
	output <= shift_register(size);
END RTL;
