-- Filename    : PermutationNetwork.vhd
-- Author:	   : Andreas Torno
-- Description : This Module describes a permutation network letting synthesis tool deside wich implementation is the appropriate

library ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;
library work;
USE work.types.all;
USE work.widths.all;

ENTITY PermutationNetwork IS
    PORT (
		-- Control ports
		clock		: IN  std_logic;
		enable		: IN  std_logic;
		reset		: IN  std_logic;
		
		-- Data ports
        in_input	: IN  std_logic_vector(PARALLELISM * WIDTH_SD - 1 DOWNTO 0);
		in_index	: IN  std_logic_vector(ROM_PERMUTATION_WIDTH - 1 downto 0);
        out_output  : OUT std_logic_vector(PARALLELISM * WIDTH_SD - 1 DOWNTO 0)
    );

END PermutationNetwork;

ARCHITECTURE RTL OF PermutationNetwork IS
TYPE sd_array is ARRAY (0 to PARALLELISM - 1) of std_logic_vector(WIDTH_SD - 1 DOWNTO 0);
TYPE ind_array is ARRAY (0 to PARALLELISM - 1) of std_logic_vector(ROM_PERMUTATION_WIDTH / PARALLELISM - 1 DOWNTO 0);
signal output		: sd_array;
signal input		: sd_array;
signal index		: ind_array;
signal output_reg	: std_logic_vector(PARALLELISM * WIDTH_SD - 1 DOWNTO 0);

BEGIN
	inputLinking: process(in_input)
	begin
		for i in 0 to (PARALLELISM - 1) LOOP
			input(i)	<= in_input((PARALLELISM - i) * WIDTH_SD - 1 downto (PARALLELISM - i - 1) * WIDTH_SD);
		end LOOP;
	end process;
	
	indexLinking: process(in_index)
	begin
		for i in 0 to (PARALLELISM - 1) LOOP
			index(i)	<= in_index((PARALLELISM - i) * ROM_PERMUTATION_WIDTH / PARALLELISM - 1 downto (PARALLELISM - i - 1) * ROM_PERMUTATION_WIDTH / PARALLELISM);
		end LOOP;
	end process;
	
	outputLinking: process(output)
	begin
		for i in 0 to (PARALLELISM - 1) LOOP
			output_reg((PARALLELISM - i) * WIDTH_SD - 1 downto (PARALLELISM - i - 1) * WIDTH_SD) <= output(i);
		end LOOP;
	end process;
	
	outputRegister: process(clock)
	begin
	if clock'event and clock='1' then
		if reset = '1' then
			out_output	<=(OTHERS => '0');
	    elsif enable = '1' then
			out_output	<= output_reg;
	    end if;
	end if;
   end process;
	
	multiplexer : for i in 0 to (PARALLELISM - 1) generate
		output(i) <= input(to_integer(unsigned(index(i))));
	end generate;

	-- permutation: process(clock)
	-- begin
	-- if clock'event and clock='1' then
		-- if reset = '1' then
			-- for i in 0 to (PARALLELISM - 1) loop
				-- output(i) <= (OTHERS => '0');
			-- end loop;
	    -- elsif enable = '1' then
			-- for i in 0 to (PARALLELISM - 1) loop
				-- output(i) <= input(to_integer(unsigned(index(i))));
			-- end loop;
	    -- end if;
	-- end if;
	-- end process;
	
	-- permutation: process(index(0), index(1), index(2), index(3), input(0), input(1), input(2), input(3))
    -- begin
		-- case index(0) is
			-- when "00"	=>	output(0) <= input(0);
			-- when "01"	=>	output(0) <= input(1);
			-- when "10"	=>	output(0) <= input(2);
			-- when "11"	=>	output(0) <= input(3);
			-- when others =>	output(0) <= (OTHERS => '0');
		-- end case;
		-- case index(1) is
			-- when "00"	=>	output(1) <= input(0);
			-- when "01"	=>	output(1) <= input(1);
			-- when "10"	=>	output(1) <= input(2);
			-- when "11"	=>	output(1) <= input(3);
			-- when others =>	output(1) <= (OTHERS => '0');
		-- end case;
		-- case index(2) is
			-- when "00"	=>	output(2) <= input(0);
			-- when "01"	=>	output(2) <= input(1);
			-- when "10"	=>	output(2) <= input(2);
			-- when "11"	=>	output(2) <= input(3);
			-- when others =>	output(2) <= (OTHERS => '0');
		-- end case;
		-- case index(3) is
			-- when "00"	=>	output(3) <= input(0);
			-- when "01"	=>	output(3) <= input(1);
			-- when "10"	=>	output(3) <= input(2);
			-- when "11"	=>	output(3) <= input(3);
			-- when others =>	output(3) <= (OTHERS => '0');
		-- end case;
    -- end process;
END RTL;
