-- Filename    : ShiftRegister.vhd
-- Author:	   : Andreas Torno
-- Description : Generic shift register with fixed length (known at the compile time)

library ieee;
USE ieee.std_logic_1164.all;
USE ieee.std_logic_arith.all;

ENTITY ShiftRegister IS
	GENERIC (
		size 		: integer := 1;
		width 		: integer := 32
	);
	
    PORT (
		-- Control ports
		clock		: IN std_logic;
		enable		: IN std_logic;
		reset		: IN std_logic;
		
		-- Data ports
        input     	: IN std_logic_vector(width-1 downto 0);
        output    	: OUT std_logic_vector(width-1 downto 0)
    );

END ShiftRegister;

ARCHITECTURE RTL OF ShiftRegister IS
	constant 	lastElement		: integer 		:= size - 2;
	type shift_type is array(lastElement downto 0) of std_logic_vector(width-1 downto 0);
	signal 		registers		: shift_type;
	signal		temp			: std_logic_vector(width-1 downto 0);
BEGIN
WRITE_THROUGH: if size = 0 generate
	output	<= input;
end generate;

FF_Register: if size = 1 generate
	FF: process(clock)
	begin
	if clock'event and clock='1' then
		if reset = '1' then
			output 	<= (OTHERS => '0');
		elsif enable = '1' then
			output	<= input;
		end if;
	end if;
	end process;
end generate;


SRL_AND_FF: if size > 1 generate
	FF: process(clock)
	begin
	if clock'event and clock='1' then
		if reset = '1' then
			output 	<= (OTHERS => '0');
		elsif enable = '1' then
			output	<= temp;
		end if;
	end if;
	end process;
	
	temp 	<= registers(lastElement);
	
	SRLUT: process(clock)
	begin
	if clock'event and clock='1' then
		if enable = '1' then
			registers(0) <= input;
			for i in 1 to lastElement LOOP
				registers(i) <= registers(i - 1);
			end LOOP;
		end if;
	end if;
	end process;
end generate;
END RTL;
